
////////////////

/*
PERCEPTION TRAINER code

navMode == "insertion" -> will insert simulated lung nodule when clicking on CXR     

Perception trainer functions:

    function predict() -> called to do lung segmentation...
    function updateGameInsertButtons() -> called when updating toggle buttons of Case assessement and lesion editor widget
    function toggleLungMask() // toggle lung mask
    function toggleLesionMask() // toggle lesions mask(s)
    function toggleLesionMarkers() // toggle lesion markers
    function toggleLesions() // toggle simulated lesions
    function reviewMarkers() // Review placed markers
    function getPTScreenshot() // Called to take screenshot to save in local storage
    function highlightLesionObject() // Called to circle lesion from on mouse over of "List of synthetic lesions"
    ....

*/
///////////////


var tensor; // holds CXR pixeldata (converted to 8 bit) to feed into tensorflow.js lung segmentation model
var predictions; // array of proabilities of pixeldata of being lung
var preds; // converted to regular JS array
var predictedImage; // predicted lung mask image
var predictSizeFactor; // size factor between segmented data and original DICOM CXRs
var meanMask; // mean of pixeldata of lung segmentation
var predictImagesDataArray; // array of DICOM pixeldata to correlate with the 128x128 segmented lung mask
var predictLocations = []; // array of x,y coordinates for possible sites to insert lung nodules
var markersList = []; // array of placed markers
// array of different colors for inserted lung nodules
var lesionColors = ['#e6194b', '#3cb44b', '#ffe119', '#4363d8', '#f58231', '#911eb4', '#46f0f0', '#f032e6', '#bcf60c', '#fabebe', '#008080', '#e6beff', '#9a6324', '#fffac8', '#800000', '#aaffc3', '#808000', '#ffd8b1', '#000075', '#808080', '#ffffff', '#000000'];


var PTeditor = false; // Should the viewer open in "assessment" or "editor" mode?
if (getParameterByName("PTeditor") == "true")
    PTeditor = true;


var saveResultToLocalStorage = true; // only save to local storage if case from case list and also not a case opened for "Case review"
if (getParameterByName("PTreview") == "true" || getParameterByName("filename") == null)
    saveResultToLocalStorage = false;

var insertObjects = new Array(); // the array to hold lesions to insert in the CXR

var caseNo = parseInt(getParameterByName("caseNo")); // which case no from "Case list"

var trainerParameters = JSON.parse(decodeURIComponent(getParameterByName("trainerParameters"))); // holds trainerParameters specified in the Pereption Trainer 
if (trainerParameters != null)
    trainerParameters = trainerParameters.map(function(e) { return parseInt(e) });

var savedInsertObjects = JSON.parse(decodeURIComponent(getParameterByName("savedInsertObjects"))); // holds settings and lesions saved in a URL
if (savedInsertObjects != null) {
    insertObjects = savedInsertObjects;
    gameMode = "plain";
}


if (gameMode == "plain")
    document.title = "Perception Trainer Viewer";

var gameModeSeed = 1; // holds seed for procedural generation
if (getParameterByName("gameModeSeed") != undefined)
    gameModeSeed = parseFloat(getParameterByName("gameModeSeed"));

var gameModeImageAugmentation = false; // augment DICOM image?
if (getParameterByName("gameModeImageAugmentation") == "true")
    gameModeImageAugmentation = true;


async function predict() { // function to invoke the tensorflow.js model to provide lung segmentation
    document.getElementById("loadCounter").innerHTML="Segmenting..."
    predictSizeFactor = [128/iWidth, 128/iHeight];
    
    
     tensor = tf.browser.fromPixels(stacks[2].images[0])
	.resizeNearestNeighbor([128,128]) // change  image size ro 128x128px
	.toFloat() // convert to float
	.div(tf.scalar(255.0)) // normalize
	.expandDims(); // expand dimensions
	
    
    predictions = await model.predict(tensor).data(); // Float32Array
    
    document.getElementById("loadCounter").innerHTML="Case generation..."; // update loading message
    
    
    preds = Array.from(predictions); // Convert to standard JS array

	// threshold the predictions
	var num;
	for (var i = 0; i < preds.length; i++) { 
		
		num = preds[i];
		
		if (num < 0.7) {   // Threshold
			preds[i] = 0;
			
		} else {
			preds[i] = 255;
			
		}
		
	}
	
	var imgData = ctx.createImageData(128, 128);
	for (var i = 0; i < imgData.data.length; i += 4) {
	  imgData.data[i+0] = 0; // Red
	  imgData.data[i+1] = 0; // Green
	  if (parseInt(preds[i/4]) == 255)
	    imgData.data[i+2] = parseInt(preds[i/4]); // Blue
	  else
	    imgData.data[i+2] = 0
	  imgData.data[i+3] = 255; // Alpha channel 
	}
	
	var canvasPredict = document.createElement("canvas");
    canvasPredict.width = 128;
    canvasPredict.height = 128;
    var ctxPredict = canvasPredict.getContext("2d");
   
    ctxPredict.putImageData(imgData, 0, 0);
	
	predictedImage = new Image();
	predictedImage.src =  canvasPredict.toDataURL("image/jpg");
	

    predictedImage.onload = function() {
        // do nothing
    }
    
    var count = 0;
    var sum = 0;
    
    // get estimate of mean of pixeldata of segmented lungs and populate predictLocations with locations for possible lung nodule insertion
    for(var b=0;b<128;b++) {
        for(var a=0;a<128;a++) {
        
            if (preds[b*128+a] == 255) {
                count++; 
                sum += predictImagesDataArray[Math.round(b/predictSizeFactor[1])*iWidth+Math.round(a/predictSizeFactor[0])];
                predictLocations.push([Math.round(a/predictSizeFactor[0]), Math.round(b/predictSizeFactor[1])]); // an array of possible coordinates for lung nodule insertion
            }
        
        }    
    }
       
   meanMask =  sum/count; // mean pixelvalue of segmented lungs
   
        // insertObjects holds parameters for lung lesions to insert
        if (trainerParameters != null) {
            var noLesions = Math.floor(Math.random() * (trainerParameters[1] - trainerParameters[0] + 1)) + trainerParameters[0];
            
       
            for(p=0;p<noLesions;p++) {
                var lesionSize = Math.floor(Math.random() * (trainerParameters[3] - trainerParameters[2] + 1)) + trainerParameters[2];
                var lesionOpacity = Math.floor(Math.random() * (trainerParameters[5] - trainerParameters[4] + 1)) + trainerParameters[4];
                var lesionMargin = Math.floor(Math.random() * (trainerParameters[7] - trainerParameters[6] + 1)) + trainerParameters[6];
               
                var random = Math.floor(Math.random() * predictLocations.length)   // insert at random location
                
                // x, y, z, lesionSize, opacity,5 = margin....10 = highlight
                insertObjects.push([predictLocations[random][0],predictLocations[random][1],1,lesionSize,lesionOpacity, lesionMargin]); 
                insertObjects[p][10] = 0;
                // insertObjects[p][11] is CNR
                insertObjects[p][12] = -1; // by default has not been marked
                
            }
            
        } // end trainerParameters != null
    
        
    
    from16to8bit2(); // called to update pixeldata with inserted lung lesions
    render();
    renderFunc();
    
    createMPR();
    
    // show either "Lesion editor widget" or "Case assessment widget"
    if (PTeditor == false)
        getGamification(); 
    else
        getObjectInserter();
    
   setTimeout(latentPlacement, 250);  // short delay in diapsying the CXR
   
}

function latentPlacement() {
    placeStacks();
    cycleStacks(firstStack);
    
}

function updateGameInsertButtons(name, on) {
    if (on == 1) {
    document.getElementById(name).style.border="2px solid black"; 
    document.getElementById(name+"2").style.border="2px solid black";  
    }
    else {
    document.getElementById(name).style.border="2px solid #818181";
    document.getElementById(name+"2").style.border="2px solid #818181";
    }
    
}

var showSegmentationMask = false;
function toggleLungMask(obj) {showSegmentationMask = !showSegmentationMask; updateStacks(); showSegmentationMask ? updateGameInsertButtons(obj.name, 1) : updateGameInsertButtons(obj.name, 0)}

var showLesionMask = false;
function toggleLesionMask(obj) {showLesionMask = !showLesionMask; updateStacks(); showLesionMask ? updateGameInsertButtons(obj.name, 1) : updateGameInsertButtons(obj.name, 0)}

var showLesionMarkers = true;
function toggleLesionMarkers(obj) {showLesionMarkers = !showLesionMarkers; updateStacks(); showLesionMarkers ? obj.style.border="2px solid black" :obj.style.border="2px solid #818181";}

var integrateLesions = true;
function toggleLesions(obj) {
    integrateLesions = !integrateLesions; 
    integrateLesions ? updateGameInsertButtons(obj.name, 1) : updateGameInsertButtons(obj.name, 0);
    from16to8bit2();
    render();
    renderFunc();
    ex3D();   
}

var caseReviewList;
function reviewMarkers(obj) {
    navMode = "";
    var correctlyIdentified = 0;
    for(var ml=0;ml<markersList.length;ml++) {
        var within = -1;;
        for(var io=0;io<insertObjects.length;io++) {
            
            dist = Math.sqrt( Math.pow(markersList[ml][0]-insertObjects[io][0],2)+Math.pow(markersList[ml][1]-insertObjects[io][1],2) );
            
            if (dist < ((insertObjects[io][3]/2)/iPixelspacing)) {
                
                within = io;
            }
            
            console.log(dist);
           
        }
        if (within >= 0) { // if marked location within bounds of lesions then list the lesion as correctly identified
            markersList[ml][3] = 1;
            insertObjects[within][12] = 1;
            correctlyIdentified++;
        }
        else {
            markersList[ml][3] = 2;
        }
        
    }
    highLightLesionObjects = true;
    
    if (caseNo > -1 && saveResultToLocalStorage) { // save results of case assessment to localstorage and notify the Perception Trainer such that "Case review" can be updated 
        if (localStorage.caseReview != undefined)
            caseReviewList = JSON.parse(localStorage.caseReview); 
        else
            caseReviewList = new Array();
        
        updateStacks();    
        
        caseReviewList.push([caseNo,getPTScreenshot(0.8), encodeURIComponent(JSON.stringify(insertObjects)), Math.abs(new Date() - dicomStartTime) / 1000, new Date()+"", [insertObjects.length, markersList.length, correctlyIdentified], gameModeSeed, gameModeImageAugmentation]); // caseNo, screenshot, insertObjects duration, date, [noNodules, tries, correctlyIdentified], gameModeSeed, gameModeImageAugmentation
        
        
        try {
            localStorage.setItem("caseReview", JSON.stringify(caseReviewList)); //saves to the database, "key", "value"
        } catch (e) {
          caseReviewList.shift(); 
          caseReviewList.shift(); 
          /// remember to ...
          localStorage.setItem("caseReview", JSON.stringify(caseReviewList));
        }
        
        
        channelPT.postMessage(["stored!"]);
    }
    
    highLightLesionObjects = false;    
    updateStacks();
    
    obj.style.display="none"; // Below code to display results of placed markers
    
    document.getElementById("segText").style.display="none";
    document.getElementById("placeMarkerButton").style.display="none";
    
    var insertHTML = "There are <b>"+insertObjects.length+"</b> synthetic lung nodules in this chest x-ray.<br>";
    
    insertHTML += "On <b>"+markersList.length+"</b> "+((markersList.length > 1 || markersList == 0) ? 'tries' : 'try') + ", ";
    insertHTML += "you correctly identified <b>"+correctlyIdentified+"</b> " +((correctlyIdentified > 1 || correctlyIdentified == 0)  ? 'nodules' : 'nodule')+ ".<br><br>"; 
    
    

    document.getElementById("afterReviewMarkers").innerHTML = insertHTML+document.getElementById("afterReviewMarkers").innerHTML+"<br><br>"; 
    
    
    insertHTML = "";
    insertHTML = "<b><u>List of synthetic lesions:</b></u><br>";
    
    // Generate "List of synthetic lesions"
    for (s=0;s<insertObjects.length;s++) {
      
      insertHTML += "<span class=\"lesionList\" onmouseout=\"javascript:highlightLesionObject("+s+",'down')\" onmouseover=\"javascript:highlightLesionObject("+s+",'high')\">"+ "<span style='color:"+lesionColors[s%lesionColors.length]+";'>"+ (s+1) +" &#10148</span> " + "X: "+insertObjects[s][0]+", "+"Y: "+insertObjects[s][1]+", "+"Diameter: ~"+insertObjects[s][3]+" mm"+", "+"Opacity: "+insertObjects[s][4]+"";
      insertHTML += ", "+"Margin: "+insertObjects[s][5];
      insertHTML += ", "+"CNR: "+insertObjects[s][11].toFixed(2)+"</span>";
      
      
      var reviewColor, reviewCharacter;
      if (insertObjects[s][12] == 1) {
        reviewColor = "#90ee90";
        reviewColor = "green";
        reviewCharacter = " <b>&#10003</b>";
      } 
      else {
        reviewColor = "red";
        reviewCharacter = " &#8722";
      }
      insertHTML += "<span style='color:"+reviewColor+"'"+">"+reviewCharacter+"</span><br>";
      
      
    }
    document.getElementById("afterReviewMarkers").innerHTML += insertHTML;
    
    
    document.getElementById("afterReviewMarkers").innerHTML += "<br><br><a href='javascript:window.close();'>Close case</>";
    
    document.getElementById("afterReviewMarkers").innerHTML += "<img style='float: right; cursor: pointer;' src='accessories/icons/iconHelp.png' border='0' title='Contrast to noise ratio = average lesion signal ROI - average background signal ROI / standard deviation of background signal ROI'>";
    
    
    document.getElementById("afterReviewMarkers").style.display="block";
    
    updateStacks();
}

var highLightLesionObjects = false;
function highlightLesionObject(no, mode) {
    
    for (s=0;s<insertObjects.length;s++) {
      if (s == no && mode == "high")
        insertObjects[s][10] = 1;
      else
        insertObjects[s][10] = 0;
    }
    console.log(no);
    updateStacks();
}
    
function getPTScreenshot(compression) { // save screenshot for local storage
    
    var tempCanvas = document.createElement('canvas');
        
    var dim = [Math.round(stacks[2].zoom*stacks[2].width), Math.round(stacks[2].zoom*stacks[2].height)];
    
    tempCanvas.width = Math.round(dim[0]/1);
    tempCanvas.height = Math.round(dim[1]/1);
    
    /*
    hidden_ctx.drawImage(
        MainCanvas,
        startClippingX,
        startClippingY,
        clippingWidth,
        clippingHeight,
        pasteX,
        pasteY,
        pasteWidth,
        pasteHeight
    );
    */
    tempCanvas.getContext('2d').drawImage(
        canvas,
        stacks[2].x,
        stacks[2].y,
        dim[0],
        dim[1],
        0,
        0,
        Math.round(dim[0]/1),
        Math.round(dim[1]/1)
    );
    
    result = tempCanvas.toDataURL('image/jpeg', compression);
    
  
   return result;    
}

function displayPerceptionTrainerGFX(i, ctx, predictedImage, offset) // 
{
    
    ctx.globalAlpha=0.08;
    if (predictedImage != undefined && showSegmentationMask) // draw overlay of segmentation mask if toggled
        ctx.drawImage(predictedImage, -offset[0],-offset[1],stacks[i].width*stacks[i].zoom,stacks[i].height*stacks[i].zoom); 
    
    if (insertObjects.length > 0) {
                 
                for(var le=0;le<insertObjects.length;le++) { // mark lesion if toggled
                    if (showLesionMask) {
                        ctx.globalAlpha = 0.18;
                        ctx.fillStyle = lesionColors[le%lesionColors.length];
                        ctx.beginPath();
                        ctx.arc(-offset[0]+insertObjects[le][0]*stacks[i].zoom,-offset[1]+insertObjects[le][1]*stacks[i].zoom, Math.floor(((insertObjects[le][3]/2)/iPixelspacing)*stacks[i].zoom), 0, 2 * Math.PI);
                        ctx.fill();
                    } // end showLesionMask
                    if (insertObjects[le][10] == 1 || highLightLesionObjects) // highlightLesion on mouse over 
                    {
                        ctx.globalAlpha = 0.9;
                        ctx.strokeStyle = lesionColors[le%lesionColors.length]; 
                        ctx.lineWidth = 4;
                        ctx.beginPath();
                        ctx.arc(-offset[0]+insertObjects[le][0]*stacks[i].zoom,-offset[1]+insertObjects[le][1]*stacks[i].zoom, Math.floor(((insertObjects[le][3]/2)/iPixelspacing)*stacks[i].zoom)*1.5, 0, 2 * Math.PI);
                        ctx.stroke();
                    }    
                    
                }
                
            }
    ctx.globalAlpha=1;        
    
}    

function displayMarkers(ctx, markersList, stacks) {
    
    if (markersList.length > 0 && showLesionMarkers) {
                        for(mL=0;mL<markersList.length;mL++) {
                            if (markersList[mL][3] == 0)
                                ctx.strokeStyle= "yellow";
                            else if (markersList[mL][3] == 1)
                                ctx.strokeStyle= "#90ee90";
                            else if (markersList[mL][3] == 2)
                                ctx.strokeStyle= "red";
                            
                            
                            var crossSize = 8;
                            ctx.beginPath();
                            ctx.moveTo(stacks[2].x+markersList[mL][0]*stacks[2].zoom-crossSize, stacks[2].y+markersList[mL][1]*stacks[2].zoom-crossSize);
                            ctx.lineTo(stacks[2].x+markersList[mL][0]*stacks[2].zoom+crossSize, stacks[2].y+markersList[mL][1]*stacks[2].zoom+crossSize);
                             ctx.moveTo(stacks[2].x+markersList[mL][0]*stacks[2].zoom+crossSize, stacks[2].y+markersList[mL][1]*stacks[2].zoom-crossSize);
                            ctx.lineTo(stacks[2].x+markersList[mL][0]*stacks[2].zoom-crossSize, stacks[2].y+markersList[mL][1]*stacks[2].zoom+crossSize);
                           
                            ctx.closePath();
                            ctx.stroke();
                            
                            
                        }
    }
    
}


function deleteObject(s) { // delete lesion from insertObjects
    
    if (s == -1)
        insertObjects = new Array();
    else
        insertObjects.splice(s, 1);
    
    from16to8bit2(1);
    render();
    renderFunc();
    ex3D();
    showObjects();
    
}

function showObjects() { // show lesions in insertObjects in "Lesion editor widget"
    
    var pane = document.getElementById('paneObjects');
    pane.innerHTML = "";
    
    for (s=0;s<insertObjects.length;s++) {
      pane.innerHTML += "<span class=\"lesionList\" onmouseout=\"javascript:highlightLesionObject("+s+",'down')\" onmouseover=\"javascript:highlightLesionObject("+s+",'high')\">"+ "<span style='color:"+lesionColors[s%lesionColors.length]+";'>"+ (s+1) +" &#10148</span> " + "X: "+insertObjects[s][0]+", "+"Y: "+insertObjects[s][1]+", "+"Diameter: ~"+insertObjects[s][3]+" mm"+", "+"Opacity: "+insertObjects[s][4]+"";
      pane.innerHTML += ", "+"Margin: "+insertObjects[s][5];
      pane.innerHTML += ", "+"CNR: "+insertObjects[s][11].toFixed(2)+"</span>";
      pane.innerHTML += " <a href='javascript:deleteObject(" + s + ")'>" + "(Delete)" + "</a><br>";
    } 
    if (insertObjects.length >= 2)
        pane.innerHTML += " <a href='javascript:deleteObject(-1)'>" + "(Delete all)" + "</a><br>";
}


function checkInsertionPlaceMarker(navMode, i) { // either insert (Editor) or place marker 
            if(navMode == "insertion") {
                // i, navMode


                var x = Math.round(stacks[i].offset.x/stacks[i].zoom);
                var y = Math.round(stacks[i].offset.y/stacks[i].zoom);
                var z = Math.round((stacks[i].imgIteration/stacks[i].slices)*stackSlices);
                
                var wantedDiameter = parseInt(document.getElementById("diameterInsertion").value);
                var wantedOpacity = parseInt(document.getElementById("opacityInsertion").value);
                var wantedMargin = parseInt(document.getElementById("marginInsertion").value);
                
                
                insertObjects.push([x,y,z,wantedDiameter,wantedOpacity,wantedMargin]);  
                insertObjects[insertObjects.length-1][10] = 0;
                   
                  
                from16to8bit2();

                render();
                renderFunc();
                ex3D();       
                showObjects();
                        
                return true;
    
           }
           
           
           if(navMode == "placeMarker") {
               // i
                var x = Math.round(stacks[i].offset.x/stacks[i].zoom);
                var y = Math.round(stacks[i].offset.y/stacks[i].zoom);
                var z = Math.round((stacks[i].imgIteration/stacks[i].slices)*stackSlices);
                markersList.push([x,y,z,0]);  
                updateStacks();   
                console.log("marker placed!");
                return true;
    
           }
           return false;
}


function CNR(lesion_Background_MeanSD) {
            var returnVal = (lesion_Background_MeanSD[0]-lesion_Background_MeanSD[2])/( Math.sqrt(0.5*(Math.pow(lesion_Background_MeanSD[1],2)+Math.pow(lesion_Background_MeanSD[3],2))));
            return Math.abs(returnVal);
}
        
function gaussianVal(height, center, sd, x) {
            return height*Math.exp(-(Math.pow(x-center,2)/(2*sd*sd)));
}


function transformImage(imgArr, xDim, yDim) { // Use TensorFlow.js functions to apply image transformations
//https://www.geeksforgeeks.org/tensorflow-js-tf-tensor4d-function/
//https://www.geeksforgeeks.org/tensorflow-js-tf-image-transform-function/
    // Specify the shape of the tensor
    var shape = [1, yDim, xDim, 1];
    // Create the tensor
    let imgTensor = tf.tensor4d(imgArr, shape);
    // Print the tensor
    imgTensor.print();
    var transformationMatrix = getTransformationMatrix(xDim, yDim);
    // Calling image.transform() method
    imgTensorTransformed = tf.image.transform(imgTensor, tf.tensor2d(
        transformationMatrix, 
        [1, 8]),'bilinear', 'constant', 0, [yDim, xDim]);
    var res = imgTensorTransformed.reshape([yDim*xDim]);
    res = res.dataSync();
    var outputArr = new Uint16Array(res);
    return outputArr;
  
}

function getTransformationMatrix(xDim, yDim) { // generate matrix for image transformation
    
    var randomForTransformationMatrix = d3.randomUniform.source(d3.randomLcg(gameModeSeed))(0, 1);
    
    let mat1 = tf.tensor2d([[1, 0,xDim/2], [0, 1,yDim/2],[0, 0,1]]); // translation to center
    let angle = (Math.PI/180)*Math.floor(randomForTransformationMatrix() * (3 - (-3) + 1) + (-3)); // angle between +/- 1-2 degress
    let mat2 =  tf.tensor2d([[Math.cos(angle), -Math.sin(angle), 0], [Math.sin(angle), Math.cos(angle), 0],[0, 0,1]]); // rotate around center
    let mat3 = tf.tensor2d([[1, 0,-xDim/2], [0, 1,-yDim/2],[0, 0,1]]); // translate back to origin
    var translationSpanPx = 5;
    let mat4 = tf.tensor2d([[1.0, 0,(xDim/100)*Math.floor(randomForTransformationMatrix() * (translationSpanPx - (-translationSpanPx) + 1) + (-translationSpanPx))], [0, 1,(yDim/100)*Math.floor(randomForTransformationMatrix() * (translationSpanPx - (-translationSpanPx) + 1) + (-translationSpanPx))],[0, 0,1]]); // additional translation
    
    result = tf.matMul(mat1, mat2);
    result = tf.matMul(result, mat3);
    result = tf.matMul(result, mat4);
    result = tf.reshape(result, [9]);
    result = result.dataSync();
    result = Array.from(result).slice(0, -1);
    return result;

}


function insertLesion(ida, maxDim, centerLes, lesRad, lesionOpacity, lesionMargin, noLesion) { // used to insert lesion in pixeldata
             // ida = imagesDataArray, centerLes = center of lesion, lesRad = radius of lesion, lesImgDim = dimensions of image to be inserted
             
             noise.seed(gameModeSeed+noLesion); // seed for Perlin noise
             
             var randomD3Uniform = d3.randomUniform.source(d3.randomLcg(gameModeSeed+noLesion))(0, 1); // additional seeded random uniform noise

             lesRad = lesRad;
             lesImgDim = lesRad*4;
             
             let lesionImage;
             
             lesionImage = [centerLes[0]-(lesImgDim/2),centerLes[1]-(lesImgDim/2),lesImgDim/2,lesImgDim/2, lesImgDim, lesImgDim, [], [],[], [],[]]; // imgX 0, imgY 1, lesX 2, lesY 3, imgW 4, imgH 5, 
              // imgData(lesion) 6, imgData(background) 7, imgData(blurred) 8, imgData(backgroundforlesion) 9, lesionsVertices 10

             lesionImage[11] = lesRad; // lesion ~ radius
             
            let tempAngle = (2*Math.PI)*randomD3Uniform(); // random angulation of nodules
                
            // generate vertices for lesion
            for(a=0;a<Math.PI*2;a=a+(Math.PI*2/360)) {
                let xoff = (Math.cos(a)+1)/2;
                let yoff = (Math.sin(a)+1)/2;
                var spike = 0;
                let r = lesionImage[11]+noise.perlin2(xoff*2,yoff*2)*(lesionImage[11]/3)+spike+(-1+randomD3Uniform()*(lesionImage[11]/6))*1;
              
                let x = r * Math.cos(a)*1.1;
                let y = r * Math.sin(a);
                let rotx = Math.cos(tempAngle)*x-Math.sin(tempAngle)*y;
                let roty = Math.sin(tempAngle)*x+Math.cos(tempAngle)*y;
        
                lesionImage[10].push([parseInt(rotx+lesionImage[2]),parseInt(roty+lesionImage[3])]);
                
            }
            
            
            let meanBackground;
            let stdBackground;
            
            let meanLesion;
            let stdLesion;
            
            
            // set backgrounds
            for(h=0;h<lesionImage[5];h++) {
                for(w=0;w<lesionImage[4];w++) {
                    // Get imgData (background)
                    lesionImage[7][h*lesionImage[4]+w] = ida[(h+lesionImage[1])*maxDim+(w+lesionImage[0])];
                    // set non-lesion background
                    lesionImage[6][h*lesionImage[4]+w] = 0;
                }
            }
            
            // get background values of region covered by lesion
            meanBackground = parseInt(math.mean(lesionImage[7]));
            stdBackground = parseInt(math.std(lesionImage[7]));
            //alert(meanBackground+" : "+stdBackground);
            
            randomD3 = d3.randomNormal.source(d3.randomLcg(gameModeSeed+noLesion))(0, windowWidth/10); // seeded normally distributed noise
            
            
            for(h=0;h<lesionImage[5];h++) {
                for(w=0;w<lesionImage[4];w++) {
                    if (checkInsidePolygon([w,h],lesionImage[10])) // if given pixel coordinate within lesion
                    {    
                        //lesionImage[6][h*lesionImage[4]+w] = 3000+meanBackground+(stdBackground*1.0)+parseInt(Math.random()*1*stdBackground);
                        dist = Math.sqrt(Math.pow(lesionImage[2]-w,2)+Math.pow(lesionImage[3]-h,2));
                        
                        // height, center, SD, dist
                        let gauss = gaussianVal(1, 0, lesRad, dist);
                        let val = gauss*(((windowWidth/20)*lesionOpacity)+0+(Math.abs(randomD3())));
                        
                        val = parseInt(val);
                        lesionImage[6][h*lesionImage[4]+w] = parseInt((val*(meanMask/meanBackground))+(lesionImage[7][h*lesionImage[4]+w]));
                    
                    }
                    else
                        lesionImage[6][h*lesionImage[4]+w] = 0;
                }
            }
            
            
             
              // Generate blurred image, imgData(lesion) 6, imgData(background) 7
              var kernelArr = setupKernels(lesionMargin*4);
              
              let lesionDensities = new Array(); 
              let lesionDensitiesBackgroundOuter = new Array();
              
              lesionImage[8] = convolutionInsertion(lesionImage[6], boxBlur,  lesionImage[4], lesionImage[5],lesionImage[7]);
              
              lesionImage[8] = convolutionInsertion(lesionImage[8], kernelArr[Math.floor(Math.random() * kernelArr.length)],  lesionImage[4], lesionImage[5],lesionImage[7]);
                 for(h=lesionImage[1];h<(lesionImage[1]+lesionImage[5]);h++) {
                    for(w=lesionImage[0];w<(lesionImage[0]+lesionImage[4]);w++) {
                       
                        
                        var blurredData = lesionImage[8][((h-lesionImage[1])*lesionImage[4])+(w-lesionImage[0])]; // blurred lesion
                        var backgroundData = lesionImage[7][((h-lesionImage[1])*lesionImage[4])+(w-lesionImage[0])]; // background
                        var originalData = lesionImage[6][((h-lesionImage[1])*lesionImage[4])+(w-lesionImage[0])]; // original lesion
                        
                            
                        
                        if (originalData <= 0 && (blurredData < backgroundData)) {// if no lesion pixel (inval3<=0) and... 
                        //if (originalData <= 0 && (blurredData <= 0)) { // use this for square delimiter around insertion 
                            lesionDensitiesBackgroundOuter.push(backgroundData);
                            continue;
                        }
                        
                        var alpha = 0.25;
                        var pixelVal = alpha*blurredData+(1-alpha)*ida[h*maxDim+w];
                        
                        
                        if (originalData <= 0) {
                            lesionDensitiesBackgroundOuter.push(pixelVal);
                            
                       
                        }
                        else
                            lesionDensities.push(pixelVal);
                        
                        
                        
                        ida[h*maxDim+w] = pixelVal; // finally set the pixel value
                       
                       
                        
                    }
                 }
                 
                 
              var lesion_Background_MeanSD = [parseInt(math.mean(lesionDensities)),parseInt(math.std(lesionDensities)),parseInt(math.mean(lesionDensitiesBackgroundOuter)),parseInt(math.std(lesionDensitiesBackgroundOuter))];
              var insertCNR = CNR(lesion_Background_MeanSD);
              insertObjects[noLesion][11] = insertCNR; // place calculated CNR in place no 11 in array
            
}

function insertSimulatedLesions(imagesDataArray,maxDim) {
    var resultTransformation;
    if (gameModeImageAugmentation && gameMode == "plain") // apply image transformation if toggled
            resultTransformation = transformImage(Array.from(imagesDataArray), maxDim, maxDim);
    else 
        resultTransformation = imagesDataArray;
    
    predictImagesDataArray = Array.from(resultTransformation); // obtain input for tensorflow.js lung segmentation predict() function
    
    if (insertObjects.length > 0 && integrateLesions) { // insert nodules if present and nodules toggled
            for(i=0;i<insertObjects.length;i++) {
               // 0, 1, 2, 3 = size, 4 = opacity
                 
                insertLesion(resultTransformation, maxDim, [insertObjects[i][0],insertObjects[i][1]], Math.floor((insertObjects[i][3]/2)/iPixelspacing), insertObjects[i][4],insertObjects[i][5], i);
            }
          
    }
    
    return resultTransformation;
    
}

